# This module is part of GitPython and is released under the
# 3-Clause BSD License: https://opensource.org/license/bsd-3-clause/

"""Initialize the Repo package."""

from .base import Repo as Repo  # noqa: F401
